-- Support for System Dynamics Model

--- Configuration variable for contributed TerraME libraries
--  @usage tmepath should point to the TerraME installation directory
tmepath = os.getenv("TME_PATH_1_3_0")
package.path = package.path .. ";"..tmepath.."/contrib/courselib/?.lua;"

local laux = require ("laux")
local show = require ("show")
local view = require ("viewer")
local ca   = require ("ca")

local abm= {}

__AgentBasedModel = {
    
    createPlacement = function (self)
        assert (self.society.max_agents_per_cell ~= nil, 
            "ABM: society should include max agents per cell")
        if self.society.placement == "entrycell" then
            assert (self.society.entrycell ~= nil, 
                "ABM: entry cell should be provided")
            self.env:createPlacement { 
                strategy = "void",
                max      = self.society.max_agents_per_cell
            }
            forEachAgent (self.society, function (agent)
                    agent:enter (self.society.entrycell)
            end)
        else   
            self.env:createPlacement { 
                strategy = self.society.placement,
                max      = self.society.max_agents_per_cell
            }
        end
    end,
    
    createMaps = function (self)
        show.createMaps (self)
    end,
    
    createAgentViews = function (self)
        show.createAgentViews (self)
    end,
    
    createGraphics = function (self)
        show.createGraphics(self)
    end,
    
    updateGraphics = function (self, ev)
        show.updateGraphics (self, ev)
    end,
    
    createTimer = function (self)
    	local timer = Timer {
            Event {time = 1, period = 1, action = 
                function(ev)
                    self.time = ev:getTime()
                    self:changes()
                    self.society:notify(ev:getTime())
                    self.cells:notify(ev:getTime())
                    if (self.views ~= nil) then 
                        self:updateGraphics(ev)
                    end
                return (self:stop() == false)
                end
               }
            }
        return timer
    end,
    
    run  = function (self, ntimes)
                self.timer:execute (ntimes)
           end
}

-- defines the metatable for the ABM Model
metaTableABMModel = {__index = __AgentBasedModel}

function abm.Model(argv)
    -- first gets the user-defined parameters
	local model = argv
    
    -- does the model define a cell space, a society and the changes function?
    assert (model.cells    ~= nil, "ABM: cell space must be provided")
    assert (model.changes  ~= nil, "ABM: model changes must be provided")
    
    -- does the cell space definition include an instance?
    if (model.cells.instance ~= nil) then
        ca.fromInstance (model.cells)
    end
    
    -- does the cell space definition include a neighborhood?
    if (model.cells.neighborhood ~= nil) then
        ca.fromNeighDef (model.cells)
    end
    
    -- initial conditions of the model (includes society definition)
    model:init()
    
    assert (model.society  ~= nil, "ABM: society must be provided")
    
    -- all other parameters are accessed via the metatable
	setmetatable(model, metaTableABMModel)
    
    -- create the Environment 
    if (model.env == nil) then 
        model.env = Environment { model.cells, model.society }
        assert (model.society.placement ~= nil, "ABM: society placement must be provided")
        model:createPlacement()
    end
            
    -- create the map for the ABM, if defined from a "map" definition
    if (model.maps ~=nil) then
        model:createMaps()
    end
    
    -- show the agents, if defined from a "agentview" definition
    if (model.agentviews ~=nil) then
        model:createAgentViews()
    end
    
    -- create the views for the model
    if (model.graphics ~= nil) then
        model:createGraphics()
    end
    
    -- is there a stopping condition?
    if (model.stop == nil) then 
        model.stop = function (self) return false end
    end
    
    -- create the timer for the model
    model.timer  = model:createTimer()
    

    
    return model
end

function abm.countNeighbors (agent, attr, value)
    local count = 0
    local cell = agent:getCell()
    forEachNeighbor(cell, function(cell, neigh)
            if (neigh:getAgent() ~= nil ) then
                local ag = neigh:getAgent()
                if ag[attr] == value then
                    count = count + 1
                end
            end
        end
    )
    return count
end

return abm