-- Support for System Dynamics Model

--- Configuration variable for contributed TerraME libraries
--  @usage tmepath should point to the TerraME installation directory
tmepath = os.getenv("TME_PATH_1_3_0")
package.path = package.path .. ";"..tmepath.."/contrib/courselib/?.lua;"

require ("constants")
local laux = require ("laux")
local show = require ("show")
local view = require ("viewer")

local ca= {}

__CellularAutomataModel = {
    
    createMaps = function (self)
        show.createMaps(self)
    end,
    
    createGraphics = function (self)
        show.createGraphics(self)
    end,
    
    updateGraphics = function (self, ev)
        show.updateGraphics (self, ev)
    end,
    
    createTimer = function (self)
    	local timer = Timer {
		Event {time = 1, period = 1, action = 
                function(ev)
                    self.time = ev:getTime()
                    self:changes()
                    self.cells:notify(ev:getTime())
                    if (self.views ~= nil) then 
                        self:updateGraphics(ev)
                    end
                end
               }
        }
        return timer
    end,
    
    run  = function (self, ntimes)
                self.timer:execute (ntimes)
           end
}

-- defines the metatable for the System Dynamics Model
metaTableCAModel = {__index = __CellularAutomataModel}

function ca.Model(argv)
    -- first gets the user-defined parameters
	local model = argv
    -- does the model define a cell space and a changes function?
    assert (model.cells    ~= nil, "Cellular Automata: cell space must be provided")
    assert (model.changes  ~= nil, "Cellular Automata: model changes must be provided") 

	-- all other parameters are accessed via the metatable
	setmetatable(model, metaTableCAModel) 
    
        -- create copies of cells based on an instance
    if (model.cells.instance ~= nil) then 
        ca.fromInstance (model.cells)
    end
    
    -- create the neighborhood based on definition 
    if (model.cells.neighborhood ~= nil) then 
        ca.fromNeighDef (model.cells)
    end
    
    -- initialize the model
    if (model.init ~= nil) then 
        model:init()
    end
    -- create the map for the CS, if defined from a "map" definition
    if (model.maps ~=nil) then
        model:createMaps()
    end
    
    -- create the views for the map
    if (model.graphics ~= nil) then
        model:createGraphics()
    end
    
    -- create the timer for the model
    model.timer  = model:createTimer()
    
    return model
end

-- initialize the cell space, based in attributes defined by an instance
function ca.fromInstance (cells)
    assert (cells.instance ~= nil) 
    
    forEachCell (cells, function (cell) 
                for k,v in pairs (cells.instance) do 
                    cell [k] = v
                end
    end)
end

-- create a neighborhood, based in attributes defined together with the cell space
function ca.fromNeighDef (cells)
    assert (cells.neighborhood ~= nil)
    assert (cells.neighborhood.strategy ~= nil)
    assert (cells.neighborhood.self ~= nil)
    assert (cells.neighborhood.wrap ~= nil)
    
    cells:createNeighborhood {
        strategy = cells.neighborhood.strategy,
        self     = cells.neighborhood.self,
        wrap     = cells.neighborhood.wrap }
end

-- count the neighbors whose attribute "attr" has the value "v"

function ca.count_neighbors(cell, attr, v)
	local count = 0
	forEachNeighbor(cell, function(cell, neigh)
		if neigh[attr] == v then
			count = count + 1
		end
	end)
	return count
end

function ca.count_past_neighbors(cell, attr, v)
	local count = 0
	forEachNeighbor(cell, function(cell, neigh)
		if neigh.past[attr] == v then
			count = count + 1
		end
	end)
	return count
end

function ca.insertPattern (cs1, cs2, x0, y0)
	for i = 0, (cs2.ydim - 1) do
		for j = 0, (cs2.xdim - 1) do
			local st2 = cs2:getCell(Coord{x = j, y = i}).state
			cs1:getCell(Coord{x=x0 + j ,y = y0 + i}).state = st2
		end
	end
end 

function ca.createRookNeighborhood(cs, len, dim)
	-- len : size of neighborhood in cells (von neumann has a size of 1)
	-- dim : dimension of the cell space
	assert (len >= 1,  "createRookNeighborhood: length is not a positive integer")
	assert (2*len < dim, "createRookNeighborhood: size is greater then cellspace dimension")
    
    cs:createNeighborhood { 
        strategy = "mxn",  
        m = len, 
        n = len,
        name = string.format ("neigh%d", len),
        weight = function (c1, c2)
                    return 1
                 end,
        filter = function (c1, c2)
                    return (c1.x == c2.x) or (c1.y == c2.y ) 
                 end
        } 
end

--[[ fills a cell space with percentage allocation of a set of classes ]]--
function ca.createRandom (cs, attr, tabClassesPer)
    local random = Random{seed = tonumber(tostring(os.time()):reverse():sub(1,6))}
    local tabPer = {}
    local tabClasses = {}
    table.insert (tabPer, 0.0)
    local i = 1
    for j = 1, #tabClassesPer do
        local tp = tabClassesPer [j]
        table.insert (tabPer, tabPer[i] + tp[2])
        table.insert (tabClasses, tp[1])
        i = i + 1
    end
    assert (tabPer[#tabPer] == 1.0, "sum of percentages does not add up to 1.0")
    forEachCell (cs, function(cell)
        local p = random:number()
        local n = #tabClasses
        for i = 1, n  do 
            if (p >= tabPer[i] and p < tabPer[i+1]) then cell[attr] = tabClasses[i] end
        end
    end)
end

--[[  find a random cell of a given state ]]--

function ca.random (cs, attr, value)
    local found = false
    local cell  = Cell{}
    repeat 
        cell = cs:sample()
        if cell[attr] == value then found = true end
    until found
    return cell
end

function ca.mapByAgentValues (cells, cell_attr, ag_attr, def)
    forEachCell (cells, function (cell)
                agent = cell:getAgent()
                if (agent == nil) then 
                    cell[cell_attr] = def
                else
                    cell[cell_attr] = agent[ag_attr]
                end
        end)
end
            
        
    

return ca