--[[ 
lauxiliary functions for lua

]]

local laux ={}
laux.version="1.0"

--[[  Parameters for the LCG random number generator 

The Lehmer random number generator, sometimes also referred to as the Park–Miller random number generator, is a variant of linear congruential generator (LCG) that operates in multiplicative group of integers modulo n. 

--]]

laux.LCGprevious   = tonumber(tostring(os.time()):reverse():sub(1,6))
laux.LCGnext       = 0
laux.LCGmodulus    = 2^31 - 1 -- Mersenne prime M31
laux.LCGmultiplier = 48271   -- ISO IEC standard for C++
laux.LCGincrement  = 0
--[[

A linear congruential generator (LCG) is an algorithm that yields a sequence of 
randomized numbers calculated with a linear equation. 
The method represents one of the oldest and best-known pseudorandom number generator algorithms
]]
function laux.randomLCG (min, max)
    laux.LCGnext = (laux.LCGmultiplier*laux.LCGprevious + laux.LCGincrement) % laux.LCGmodulus 
	local rand = (laux.LCGnext % (max - min + 1)) + min  
	laux.LCGprevious = laux.LCGnext
	return rand
end

--- A random value from a array in lua
function laux.randomValueFromArray (array)
    return array [ laux.randomLCG (1, #array)]
end

--[[
	Wait some time before the next iteration 
	to help visualize the model better
]]
function laux.wait(n)
	for i = 0, n*100000 do io.write("") end -- do nothing 100.000 times
end

--[[
    Find the type of operating system (unixes or windows)
]]-- 

function laux.getos()
        -- Unix, Linux varients
        sep = package.config:sub(1, 1)
        if sep == "/" then 
            return "unix" 
        else 
            return "windows"
        end
end
--[[
    Find the maximum of two values
]]-- 

function laux.max (x,y)
    assert (x ~= nil)
    assert (y ~= nil)
    return (x > y) and x or y
end

--[[
    Find the minumum of two values
]]-- 

function laux.max (x,y)
    assert (x ~= nil)
    assert (y ~= nil)
    return (x > y) and x or y
end

--[[
    Clone a lua table
]]-- 
function laux.clone (luaTable)

	local newTable = {}
	for key,val in pairs(luaTable) do
		if type(val) == 'table' then
			val = laux.clone (val)
		end
		newTable [key] = val
	end
	return newTable
end

--[[
  print values of a lua Table
]]
function laux.printTab (t)
    for k,v in pairs(t) do
        if type(v) == 'table' then
            laux.printTab (v)
        end
        print (k,v)
    end
end

--[[
  print values of a lua Table with numbers 
]]
function laux.printNumTab (tab)
    local out = out or ""
    for k,v in pairs (tab) do
        if type(v) == 'table' then
            laux.printNumTab (v)
        else
            out = out .. v .. " "
        end
    end
    out = "{" .. out .. "}"
    print (out)
end

--[[ 
	Find out if an element is one of the values of a table
]]
function laux.isOneOf(tab, elem)
  	for k, v in ipairs(tab) do
    	if v == elem then
      	return true
    	end
  	end
  	return false
end

--[[
  Count the occurences of an element in a table
]]
function laux.count(tab, elem)
    local count
    count = 0
    for k,v in pairs(tab) do
        if elem == v then count = count + 1 end
    end
    return count
end

--[[
   Converts an integer to binary
   Returns a table of bits, most significant first.
]]--
function laux.to8Bits(num)
    -- returns a table of bits, most significant first.
    bits = 8
    local t={} -- will contain the bits        
    for b=bits,1,-1 do
        t[b]=math.fmod(num,2)
        num=(num-t[b])/2
    end
    local rule = {}
        rule["111"] = t[1]
        rule["110"] = t[2]
        rule["101"] = t[3]
        rule["100"] = t[4]
        rule["011"] = t[5]
        rule["010"] = t[6]
        rule["001"] = t[7]
        rule["000"] = t[8]
    return rule 
end
return laux