--- Graphics library for TerraME
--  @author Gilberto Camara, 2014

--- Configuration variable for contributed TerraME libraries
--  @usage tmepath should point to the TerraME installation directory
tmepath = os.getenv("TME_PATH_1_3_0")
package.path = package.path .. ";"..tmepath.."/contrib/courselib/?.lua;"

--- Dependencies of the graphics library
--  @usage ding calls the dingbats module
local view = require ("viewer")
local laux = require ("laux")

local show = {}

function show.createMaps (model)
    assert (model.maps ~=nil, "map definitions not provided")
    -- create a first observer (Windows bug)
    if (model.firstobserver == nil) then
        show.createFirstObserver ()
        model.firstobserver = true
    end
    
    -- create the themes to be visualised
    for viewname, viewdef in pairs (model.maps) do
        -- map views are composed of one of or more themes
        -- to be visualised
        -- each view should have at least one theme
        local mapdef = {
            attributes = {},
            legends    = {}
        }
        for _,theme in pairs (viewdef) do
            assert (theme.attribute ~= nil, 
                    "each map theme should have at least one attribute")
            -- create a legend for this theme
            local tme_legend = view.Legend (theme)
            table.insert (mapdef.attributes, theme.attribute)
            table.insert (mapdef.legends, tme_legend)      
        end
        local obsname = "obs_"..viewname
        model.obsname  = view.Map {
            subject    = model.cells,
            attributes = mapdef.attributes,
            legends    = mapdef.legends
        }

    end
end
function show.createAgentViews (model)
    assert (model.agentviews ~= nil, "agent view definitions must be provided")        
        
    -- create the themes to be visualised
    for _,viewdef in pairs (model.agentviews) do
        -- map views are composed of one of or more themes
        -- to be visualised
        -- each view should have at least one theme
        for tn,theme in pairs (viewdef) do
            assert (theme.attribute ~= nil, 
                    "each symbol should have at least one attribute")
            assert (theme.overlay ~= nil, 
                    "to display symbols, please inform the theme to be overlaid")
            local obsname = "obs_"..theme.overlay
            assert (model.obsname ~= nil, 
                    "map where symbols will be overlaid has not been defined")
            -- table that will contain the symbol definitions
            local symdef = {
                attributes = {}
            }
            -- insert the attributes name into a table
            table.insert (symdef.attributes, theme.attribute)
            -- the subject of the agent view
            symdef.subject = model.society
            symdef.observer = model.obsname
            for k,v in pairs (theme) do
                symdef[k] = v
            end      
        
            local obsname = "obs_agent_"..tn

            model.obsname  = view.Symbol (symdef)
        end
    end
end
function show.createGraphics (model)
    assert (model.graphics ~= nil, 
        "model must have a table defining which graphics will be displayed")
    
    if model.views == nil then 
       model.views = {}
    end
    
    if (model.firstobserver == nil) then
        show.createFirstObserver ()
        model.firstobserver = true
    end

    for k, v in pairs (model.graphics) do
        if k == "timeseries" then 
            show.createTimeSeriesPlots(model, v)
        elseif k == "phasespace" then
            show.createPhaseSpacePlots (model, v)
        elseif k == "cobweb" then
            show.createCobwebPlots (model, v)
        end
    end
end
    
function show.createFirstObserver()
    if (laux.getos() == "windows") then
        local subj = {}
        subj ["x"] = 0.0
        local first = view.Plot {
                subject     = Cell (subj),    
                attributes = {"x"}
        }
        first.subject:notify(0)
    end
end
    
function show.createTimeSeriesPlots (model, tsTable)
    for _,tab in pairs (tsTable) do
        local subj = {}
        local atts = {}
        for _, name in pairs (tab) do
            subj [name] = model [name]
            table.insert (atts, name) 
        end
        local plot = view.Plot {
                subject    = Cell (subj),    
                attributes = atts,
                xLabel     = "time"
            }
        plot.subject:notify(0)
        plot.type = "timeseries"
        table.insert (model.views, plot)
    end
end
    
function show.createPhaseSpacePlots (model, psTable)
    for _, tab in pairs (psTable) do
        assert (#tab >= 2, "phase space plots have at least two variables")
        local subj = {}
        x = tab[1]
        subj [x] = model [x]
        
        local atts = {}
        for i = 2, #tab do
            y = tab[i]
            subj [y] = model [y]
            table.insert (atts, y)
        end
        
        local plot = view.Plot {
            subject    = Cell (subj),    
            attributes = atts,
            title      = "Phase Space Plot",
            xAxis      = x, 
            xLabel     = x,
            yLabel     = y                      
        }
        plot.xVar = x
        plot.type = "phasespace"
        plot.subject:notify(0) 
        table.insert (model.views, plot)
    end 
end   
    
function show.createCobwebPlots (model, cbTable)
    for _, tab in pairs (cbTable) do
        assert (#tab == 1, "cobweb plots use a single variable")
            
        for _, name in pairs (tab) do
            local st = { xp = 0.0, fx = model[name], yline = 0.0}
            local plot = view.Plot {
                subject    = Cell (st),    
                attributes = {"fx", "yline"},
                title      = "Cobweb plot",
                xAxis      = "xp", 
                xLabel     = "xp",
                yLabel     = "fx"                       
            }
            plot.xVar = name
            plot.type = "cobweb"
            plot.subject:notify(0) 
            table.insert (model.views, plot)
        end 
    end
end
    
function show.updateGraphics (model, ev)
    for _,v in pairs (model.views) do
        if v.type == "timeseries" or v.type == "phasespace" then
            show.updateTimeSeries (model, v, ev)
        elseif v.type == "cobweb" then 
            show.updateCobwebPlot (model, v, ev)
        end
    end
end
    
function show.updateTimeSeries (model, v, ev)
    for _,name in pairs (v.attributes) do
        v.subject [name] = model [name]
    end
    v.subject:notify(ev:getTime())
end
        
function show.updateCobwebPlot (model, v, ev)
        --[[ Cobweb plot
        Starting from an initial point (x0, f(x0)), repeat two steps:
        1. Plot horizontally across from this point to the diagonal line. 
           This has the coordinates (f(x), f(x)).
        2. Plot vertically from the point on the diagonal to the function curve.
           This has the coordinates (f(x), f(f(x))).
        --]]
        -- plot step 1 of the cobweb plot
        v.subject["xp"] = v.subject["fx"]
        v.subject["fx"] = v.subject["fx"]
        
        -- plot the x = y line
        v.subject["yline"] = v.subject["xp"]
            
        v.subject:notify(ev:getTime())
        
        -- plot step 2 of the cobweb plot
        v.subject["xp"] = v.subject["fx"]
        v.subject["fx"] = model[v.xVar]
        
        v.subject:notify(ev:getTime())

end

return show